package jmetal.operators.crossover;

import jmetal.core.Solution;
import jmetal.encodings.solutionType.ArrayRealSolutionType;
import jmetal.encodings.solutionType.RealSolutionType;
import jmetal.util.Configuration;
import jmetal.util.JMException;
import jmetal.util.PseudoRandom;
import jmetal.util.wrapper.XReal;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Scanner;

/**
 * This class allows to apply a SBX crossover operator using two parent
 * solutions.
 */
public class ArithmeticalCrossover extends Crossover {

	private Double crossoverProbability_ = 0.9;
	private Double lambda_ = 0.2;

	/**
	 * Valid solution types to apply this operator
	 */
	private static final List VALID_TYPES = Arrays.asList(
			RealSolutionType.class, ArrayRealSolutionType.class);

	/**
	 * Constructor Create a new SBX crossover operator whit a default index
	 * given by <code>DEFAULT_INDEX_CROSSOVER</code>
	 */
	public ArithmeticalCrossover(HashMap<String, Object> parameters) {
		super(parameters);

		if (parameters.get("probability") != null)
			crossoverProbability_ = (Double) parameters.get("probability");
		
		if (parameters.get("lambda") != null)
			crossoverProbability_ = (Double) parameters.get("lambda");

	} 

	/**
	 * Perform the crossover operation.
	 * 
	 * @param probability
	 *            Crossover probability
	 * @param parent1
	 *            The first parent
	 * @param parent2
	 *            The second parent
	 * @return An array containing the two offsprings
	 */
	public Solution[] doCrossover(double probability, Solution parent1,
			Solution parent2) throws JMException {

		Solution[] offSpring = new Solution[2];

		offSpring[0] = new Solution(parent1);
		offSpring[1] = new Solution(parent2);

	
		XReal x1 = new XReal(parent1);
		XReal x2 = new XReal(parent2);
		XReal offs1 = new XReal(offSpring[0]);
		XReal offs2 = new XReal(offSpring[1]);

	
		int numberOfVariables = x1.getNumberOfDecisionVariables();
		
		if (PseudoRandom.randDouble() <= probability) {
			for (int i = 0; i < numberOfVariables; i++) {

				double p1 = x1.getValue(i);
				double p2 = x2.getValue(i);
				
				if (PseudoRandom.randDouble() > 0.5 || p1 == p2)
					continue;
				
			
				double yL = x1.getLowerBound(i);
				double yU = x1.getUpperBound(i);
				
				double y2 = Math.max(p1, p2);
				double y1 = Math.min(p1, p2);
				
				double rand = PseudoRandom.randDouble();
				double beta = 1.0 + (2.0 * (y1 - yL) / (y2 - y1));
				double alpha = 2.0 - java.lang.Math.pow(beta,
						-(31));
				
	
				double betaq;
				if (rand <= (1.0 / alpha)) {
					betaq = java.lang.Math.pow((rand * alpha),
							(1.0 / 31.0));
				} else {
					betaq = java.lang.Math.pow((1.0 / (2.0 - rand
							* alpha)),
							(1.0 / 31.0));
				} // if
				
				
			
				double c1 = 0.5 * (1 + betaq) * p1 + 0.5 * (1 - betaq) * p2;
				
		
			
			
	//			double u = PseudoRandom.randDouble(0.99, 1.01);
	//			c1 = u * y1 + (1-u)* y2;
	//			c1 = u * p1 + (1-u) * p2;
				
/*				System.out.print("a: " + 0.5 * (1 + betaq) + "\t");
				System.out.println("b: " + 0.5 * (1 - betaq));*/
				
				beta = 1.0 + (2.0 * (yU - y2) / (y2 - y1));
				alpha = 2.0 - java.lang.Math.pow(beta,
						- 31.0);

				if (rand <= (1.0 / alpha)) {
					betaq = java.lang.Math.pow((rand * alpha),
							(1.0 / 31.0));
				} else {
					betaq = java.lang.Math.pow((1.0 / (2.0 - rand
							* alpha)),
							(1.0 / 31.0));
				} // if

			
				double c2 = 0.5 * (1 - betaq) * p1 + 0.5 * (1 + betaq) * p2;
				
/*				c1 = y1;
				c2 = yL + yU - y1;*/
				

				
//				u = PseudoRandom.randDouble(0.99, 1.01);
				
				
				
		//		c2 = (1 - u) * y1 + u * y2;
		//		c2 = (1 - u) * p1 + u * p2;
				
	/*			System.out.print("a1: " + 0.5 * (1 + betaq) + "\t");
				System.out.println("b1: " + 0.5 * (1 - betaq));*/
		
			/*	double p1 = x1.getValue(i);
				double p2 = x2.getValue(i);
				
				double u = PseudoRandom.randDouble(0.99, 1.01);
				//	u = 0.99;
				double c1 = u * p1 + (1-u)* p2;
				double c2 =  u * p2 + (1-u)* p1;
				
				double yL = x1.getLowerBound(i);
				double yU = x1.getUpperBound(i);*/
					
				if (c1 < yL)
					c1 = yL;
				else if (c1 > yU)
					c1 = yU;

				if (c2 < yL)
					c2 = yL;
				else if (c2 > yU)
					c2 = yU;
				
			
				
/*				offs1.setValue(i, c1);
				offs2.setValue(i, c2);*/
	
				if (PseudoRandom.randDouble() <= 0.5) {
					offs1.setValue(i, c1);
					offs2.setValue(i, c2);
				}
				else {
					offs1.setValue(i, c2);
					offs2.setValue(i, c1);
				}
				
			
			}
		}
		
		return offSpring;
	} // doCrossover

	/**
	 * Executes the operation
	 * 
	 * @param object
	 *            An object containing an array of two parents
	 * @return An object containing the offSprings
	 */
	public Object execute(Object object) throws JMException {
		Solution[] parents = (Solution[]) object;

		if (parents.length != 2) {
			Configuration.logger_
					.severe("UniformCrossover.execute: operator needs two "
							+ "parents");
			Class cls = java.lang.String.class;
			String name = cls.getName();
			throw new JMException("Exception in " + name + ".execute()");
		} // if

		if (!(VALID_TYPES.contains(parents[0].getType().getClass()) && VALID_TYPES
				.contains(parents[1].getType().getClass()))) {
			Configuration.logger_.severe("ArithmeticalCrossover.execute: the solutions "
					+ "type " + parents[0].getType()
					+ " is not allowed with this operator");

			Class cls = java.lang.String.class;
			String name = cls.getName();
			throw new JMException("Exception in " + name + ".execute()");
		} // if

		Solution[] offSpring;
		offSpring = doCrossover(crossoverProbability_, parents[0], parents[1]);

		// for (int i = 0; i < offSpring.length; i++)
		// {
		// offSpring[i].setCrowdingDistance(0.0);
		// offSpring[i].setRank(0);
		// }
		return offSpring;
	} // execute
} // SBXCrossover
